/*
 * INCLUDE FILE FOR REFLECTA.C
 */

#include "reflecta_scancmd.h"
#include "reflecta_usb.h"


/* =========================================================================
 * 
 * Reflecta scanner commands
 * 
 * ========================================================================= */

/* Standard SCSI command codes */
#define SCSI_TEST_UNIT_READY    0x00
#define SCSI_REQUEST_SENSE      0x03
#define SCSI_READ               0x08
#define SCSI_WRITE              0x0A
#define SCSI_PARAM              0x0F
#define SCSI_INQUIRY            0x12
#define SCSI_MODE_SELECT        0x15
#define SCSI_COPY               0x18
#define SCSI_SCAN               0x1B

/* Non-standard SCSI command codes */
#define SCSI_SET_SCAN_HEAD      0xD2
#define SCSI_READ_GAIN_OFFSET   0xD7
#define SCSI_WRITE_GAIN_OFFSET  0xDC
#define SCSI_READ_STATE         0xDD

/* Additional SCSI READ/WRITE codes */
#define SCSI_HALFTONE_PATTERN   0x11
#define SCSI_SCAN_FRAME         0x12

/**
 * Perform a TEST UNIT READY (SCSI command code 0x00)
 * Returns status->sane_status:
 * - SANE_STATUS_GOOD if device is ready
 * - SANE_STATUS_DEVICE_BUSY if device is still busy after repeat tries
 * - SANE_STATUS_CHECK_CONDITION with accompanying sense codes if command
 *   returned a CHECK CONDITION
 * - other SANE status code if TEST UNIT READY failed or if it returned
 *   CHECK CONDITION and REQUEST SENSE failed
 * 
 * @param device_number Device number
 * @return SANE_status SANE_STATUS_GOOD if ready, SANE_STATUS_DEVICE_BUSY if not
 */
void cmdIsUnitReady(SANE_Int device_number, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    SANE_Byte command[6];
    
    DBG (DBG_sane_info, "cmdIsUnitReady()\n");

    setCommand(command,SCSI_TEST_UNIT_READY,0);

    commandScannerRepeat(device_number, command, NULL, 0, status, repeat);

    DBG (DBG_sane_info, "cmdIsUnitReady() return status = %s\n",sane_strstatus(status->sane_status));
}

/**
 * Perform a REQUEST SENSE (SCSI command code 0x03)
 * Returns status->sane_status:
 * - SANE_STATUS_GOOD is the command executes OK
 * - other SANE status code if REQUEST SENSE fails
 * The sense fields in status are always 0. A REQUEST SENSE is not repeated if
 * the device returns SANE_STATUS_DEVICE_BUSY.
 * 
 * @param device_number Device number
 * @param sense Sense data
 * @param status Command result status
 * @see struc Reflecta_Sense
 */
void cmdGetSense(SANE_Int device_number, struct Reflecta_Sense* sense, struct Reflecta_Command_Status *status)
{
    SANE_Byte command[6];
    SANE_Int size = 14;
    SANE_Byte data[size];
    
    DBG (DBG_sane_info, "cmdGetSense()\n");
    
    setCommand(command,SCSI_REQUEST_SENSE,size);
    
    memset(data,'\0',size);
    commandScanner(device_number, command, data, size, status);
    if (status->sane_status != SANE_STATUS_GOOD) {
        return;
    }
    
    /* Decode data recieved */
    sense->errorCode = getByte(data,0);
    sense->segment = getByte(data,1);
    sense->senseKey = getByte(data,2);
    getBytes(sense->info,data,3,4);
    sense->addLength = getByte(data,7);
    getBytes(sense->cmdInfo,data,8,4);
    sense->senseCode = getByte(data,12);
    sense->senseQualifier = getByte(data,13);   
    status->sane_status = SANE_STATUS_GOOD;
}

/**
 * Read the halftone pattern with the specified index. This requires two
 * commands, one to ask the device to prepare the pattern, and one to read it.
 * 
 * @param device_number Device number
 * @param index index of halftone pattern
 * @param pattern Halftone pattern (not implemented)
 * @return SANE_status
 * @see Reflecta_Halftone_Pattern
 */
void cmdGetHalftonePattern(SANE_Int device_number, SANE_Int index, struct Reflecta_Halftone_Pattern* pattern, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    SANE_Byte command[6];
    SANE_Int size = 256; /* Assumed maximum pattern size */
    SANE_Byte data[size];
    int k;
    
    DBG (DBG_sane_info, "cmdGetHalftonePattern()\n");
    
    /* Ask scanner to prepare the pattern with the given index. Only 6 bytes of data. */
    setCommand(command,SCSI_WRITE,6);
    memset(data,'\0',6);
    data[0] = SCSI_HALFTONE_PATTERN | 0x80; /* set bit 7 means prepare read */
    data[4] = index;
    
    commandScanner(device_number, command, data, 6, status);
    if (status->sane_status != SANE_STATUS_GOOD) {
        return;
    }
    
    /* Read pattern */
    setCommand(command,SCSI_READ,size);
    
    memset(data,'\0',size);
    commandScanner(device_number, command, data, size, status);
    
    /*TODO: analyse */
    fprintf(stderr, "Halftone pattern %d:\n",index);
    int psize = data[3];
    psize = (psize<<8) + data[2];
    SANE_Char* desc = data+4+psize;
    data[4+psize+16] = '\0';
    fprintf(stderr,"Descr. offset from byte 4 = %d, %16s, index = %d, size = %dx%d\n",psize,desc,data[4]&0x7F,data[6],data[7]);
}

/**
 * Read the scan frame with the specified index. This requires two
 * commands, one to ask the device to prepare the pattern, and one to read it.
 * 
 * @param device_number Device number
 * @param frame Scan frame
 * @return SANE_status
 * @see Reflecta_Scan_Frame
 */
void cmdGetScanFrame(SANE_Int device_number, SANE_Int index, struct Reflecta_Scan_Frame* frame, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    SANE_Byte command[6];
    SANE_Int size = 256; /* Assumed maximum frame size */
    SANE_Byte data[size];
    int k;
    
    DBG (DBG_sane_info, "cmdGetScanFrame()\n");
    
    /* Ask scanner to prepare the scan frame with the given index. Only 6 bytes of data. */
    setCommand(command,SCSI_WRITE,6);
    memset(data,'\0',6);
    data[0] = SCSI_SCAN_FRAME | 0x80; /* set bit 7 means prepare read */
    data[4] = index;
    
    commandScanner(device_number, command, data, 6, status);
    if (status->sane_status != SANE_STATUS_GOOD) {
        return;
    }
    
    /* Read scan frame */
    setCommand(command,SCSI_READ,size);
    
    memset(data,'\0',size);
    commandScanner(device_number, command, data, size, status);

    /* Decode data */
    frame->code = getByte(data,0);
    frame->size = getShort(data,2);
    frame->index = getByte(data,4);
    frame->x0 = getShort(data,6);
    frame->y0 = getShort(data,8);
    frame->x1 = getShort(data,10);
    frame->y1 = getShort(data,12);

    DBG (DBG_info, "cmdGetScanFrame() set:\n");
    DBG (DBG_info, " x0,y0 = %d,%d\n",frame->x0,frame->y0);
    DBG (DBG_info, " x1,y1 = %d,%d\n",frame->x1,frame->y1);
    DBG (DBG_info, " code = %d\n",frame->code);
    DBG (DBG_info, " index = %d\n",frame->index);
    DBG (DBG_info, " size = %d\n",frame->size);
}

/**
 * Read the relative exposure time for the specified colorbits. This requires two
 * commands, one to ask the device to prepare the value, and one to read it.
 * 
 * @param device_number Device number
 * @param time Relative exposure time(s)
 * @return SANE_status
 * @see Reflecta_Exposure_Time
 */
void cmdGetRelativeExposureTime(SANE_Int device_number, SANE_Int colorbits, struct Reflecta_Exposure_Time* time, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    DBG (DBG_sane_info, "cmdGetRelativeExposureTime(): not implemented\n");
    status->sane_status = SANE_STATUS_INVAL;
}

/**
 * Read the highlight and shadow levels with the specified colorbits. This requires two
 * commands, one to ask the device to prepare the value, and one to read it.
 * 
 * @param device_number Device number
 * @param hgltshdw Highlight and shadow level(s)
 * @return SANE_status
 * @see Reflecta_Highlight_Shadow
 */
void cmdGetHighlightShadow(SANE_Int device_number, SANE_Int colorbits, struct Reflecta_Highlight_Shadow* hgltshdw, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    DBG (DBG_sane_info, "cmdGetHighlightShadow(): not implemented\n");
    status->sane_status = SANE_STATUS_INVAL;
}

/**
 * Read the shading data parameters with the specified index. This requires two
 * commands, one to ask the device to prepare the value, and one to read it.
 * 
 * @param device_number Device number
 * @param shading Shading data parameters
 * @return SANE_status
 * @see Reflecta_Shading_Parameters
 */
void cmdGetShadingParameters(SANE_Int device_number, SANE_Int index, struct Reflecta_Shading_Parameters* shading, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    DBG (DBG_sane_info, "cmdGetShadingParameters(): not implemented\n");
    status->sane_status = SANE_STATUS_INVAL;
}

/**
 * Read scanned data from the scanner memory into a byte array. The lines
 * argument specifies how many lines will be read, the size argument specifies
 * the total amount of bytes in these lines. Use cmdGetScanParameters() to
 * determine the current line size and the number of available lines.\n
 * If there is scanned data available, it should be read. Waiting too long
 * causes the scan to stop, probably because a buffer is filled to its limits
 * (if so, it is approximately 2Mb in size). I haven't tried what happens if you
 * start reading after a stop. Reading to fast causes the scanner to return
 * a busy status, which is not a problem.
 * This is a SCSI READ command (code 0x08). It is distinguished from the other
 * READ commands by the context in which it is issued: see cmdStartScan().
 *
 * @param device_number
 * @param data
 * @param lines
 * @param size
 * @return SANE_status
 */
void cmdGetScannedLines(SANE_Int device_number, SANE_Byte* data, SANE_Int lines, SANE_Int size, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    SANE_Byte command[6];

    DBG (DBG_sane_info, "cmdGetScannedLines(): %d (%d bytes)\n",lines,size);
    
    setCommand(command,SCSI_READ,lines);    
    memset(data,'\0',size);
    
    commandScannerRepeat(device_number, command, data, size, status, repeat);
}

/**
 * Set the halftone pattern with the given index to the specified pattern. The
 * command is a SCSI WRITE command (code 0x0A, write code 0x11).
 * 
 * @param device_number Device number
 * @param index Pattern index (0-7)
 * @param pattern Halftone pattern (not implemented)
 * @return SANE_status
 * @see Reflecta_Halftone_Pattern
 */
void cmdSetHalftonePattern(SANE_Int device_number, SANE_Int index, struct Reflecta_Halftone_Pattern* pattern, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    DBG (DBG_sane_info, "cmdSetHalftonePattern(): not implemented\n");
    status->sane_status = SANE_STATUS_INVAL;
}

/**
 * Set the scan frame with the given index to the frame. The command is a SCSI
 * WRITE command (code 0x0A, write code 0x12).
 * 
 * @param device_number Device number
 * @param index Frame index (0-7)
 * @param frame Scan frame
 * @return SANE_status
 * @see Reflecta_Scan_Frame
 */
void cmdSetScanFrame(SANE_Int device_number, SANE_Int index, struct Reflecta_Scan_Frame* frame, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    SANE_Byte command[6];
    SANE_Int size = 14;
    SANE_Byte data[size];
    
    DBG (DBG_sane_info, "cmdSetScanFrame()\n");
    
    setCommand(command,SCSI_WRITE,size);
    
    DBG (DBG_info, "cmdSetScanFrame() set:\n");
    DBG (DBG_info, " x0,y0 = %d,%d\n",frame->x0,frame->y0);
    DBG (DBG_info, " x1,y1 = %d,%d\n",frame->x1,frame->y1);
    DBG (DBG_info, " code = %d\n",frame->code);
    DBG (DBG_info, " index = %d\n",frame->index);
    DBG (DBG_info, " size = %d\n",frame->size);

    /* Code data */
    memset(data,'\0',size);
    setShort(SCSI_SCAN_FRAME,data,0);
    setShort(size-4,data,2); /* size: one frame, 5 shorts */
    setShort(index,data,4);
    setShort(frame->x0,data,6);
    setShort(frame->y0,data,8);
    setShort(frame->x1,data,10);
    setShort(frame->y1,data,12);
    
    commandScannerRepeat(device_number, command, data, size, status, repeat);
}

/**
 * Set the relative exposure time to the given values. Only the first
 * Reflecta_Exposure_Time_Color is used. The command is a SCSI
 * WRITE command (code 0x0A, write code 0x13).
 * 
 * @param device_number Device number
 * @param time Relative exposure time
 * @return SANE_status
 * @see Reflecta_Exposure_Time
 */
void cmdSetRelativeExposureTime(SANE_Int device_number, struct Reflecta_Exposure_Time* time, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    DBG (DBG_sane_info, "cmdSetRelativeExposureTime(): not implemented\n");
    status->sane_status = SANE_STATUS_INVAL;
}

/**
 * Set the highlight and shadow levels to the given values. Only the first
 * Reflecta_Highlight_Shadow_Color is used. The command is a SCSI
 * WRITE command (code 0x0A, write code 0x14).
 * 
 * @param device_number Device number
 * @param hgltshdw highlight and shadow level
 * @return SANE_status
 * @see Reflecta_Highlight_Shadow
 */
void cmdSetHighlightShadow(SANE_Int device_number, struct Reflecta_Highlight_Shadow* hgltshdw, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    DBG (DBG_sane_info, "cmdSetHighlightShadow(): not implemented\n");
    status->sane_status = SANE_STATUS_INVAL;
}

/**
 * Set the CCD-mask for the colors set in the given color bit mask. The mask
 * array must contain 2x5340 = 10680 bytes. The command is a SCSI WRITE command
 * (code 0x0A, write code 0x16). 
 * (The command is able to handle more masks at once, but that is not implemented.)
 * 
 * @param device_number Device number
 * @param colorbits 0000RGB0 color bit mask; at least one color bit must be set
 * @param mask CCD mask to use
 * @return SANE_status
 */
void cmdSetCCDMask(SANE_Int device_number, SANE_Byte colorbits, SANE_Byte* mask, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    DBG (DBG_sane_info, "cmdSetCCDMask(): not implemented\n");
    status->sane_status = SANE_STATUS_INVAL;
}

/* SCSI PARAM, code 0x0F */
/**
 * Get the parameters of an executed scan, such as width, lines and bytes, which
 * are needed to calculate the parameters of the READ-commands which read the
 * actual scan data.
 *
 * @param device_number Device number
 * @param parameters Scan parameters
 * @return SANE_status
 * @see Reflecta_Scan_Parameters
 */
void cmdGetScanParameters(SANE_Int device_number, struct Reflecta_Scan_Parameters* parameters, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    SANE_Byte command[6];
    SANE_Int size = 18;
    SANE_Byte data[size];
    
    DBG (DBG_info, "cmdGetScanParameters()\n");
    
    setCommand(command,SCSI_PARAM,size);
    memset(data,'\0',size);
    
    commandScannerRepeat(device_number, command, data, size, status, repeat);
    if (status->sane_status != SANE_STATUS_GOOD) {
        return;
    }
    
    /* Decode data recieved */
    parameters->width = getShort(data,0);
    parameters->lines = getShort(data,2);
    parameters->bytes = getShort(data,4);
    parameters->filterOffset1 = getByte(data,6);
    parameters->filterOffset2 = getByte(data,7);
    parameters->period = getInt(data,8);
    parameters->scsiTransferRate = getShort(data,12);
    parameters->availableLines = getShort(data,14);
    
    DBG (DBG_info, "cmdGetScanParameters() read:\n");
    DBG (DBG_info, " width = %d\n",parameters->width);
    DBG (DBG_info, " lines = %d\n",parameters->lines);
    DBG (DBG_info, " bytes = %d\n",parameters->bytes);
    DBG (DBG_info, " offset1 = %d\n",parameters->filterOffset1);
    DBG (DBG_info, " offset2 = %d\n",parameters->filterOffset2);
    DBG (DBG_info, " available lines = %d\n",parameters->availableLines);
}

/**
 * Read INQUIRY block from device (SCSI command code 0x12). This block contains
 * information about the properties of the scanner.
 * Returns status->sane_status:
 * - SANE_STATUS_GOOD if the INQUIRY command succeeded
 * - SANE_STATUS_DEVICE_BUSY if device is busy after repeat retries
 * - SANE_STATUS_CHECK_CONDITION with accompanying sense codes if the command
 *   returned a CHECK CONDITION
 * - other SANE status code if INQUIRY failed or if it returned CHECK CONDITION
 *   and REQUEST SENSE failed
 * 
 * @param device_number Device number
 * @param data Input or output data buffer
 * @param size Size of the data buffer
 * @return SANE_status
 * @see Reflecta_Scanner_Properties
 */
void cmdDoInquiry(SANE_Int device_number, struct Reflecta_Scanner_Properties* inq, SANE_Byte size, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    SANE_Byte command[6];
    SANE_Byte data[256];
    int k;
    
    DBG (DBG_sane_info, "cmdGetScannerProperties()\n");
    
    setCommand(command,SCSI_INQUIRY,size);
    memset(data,'\0',256); /* size may be less than 256, so prevent returning noise */
    
    commandScannerRepeat(device_number, command, data, size, status, repeat);
    if (status->sane_status != SANE_STATUS_GOOD) {
        return;
    }
    
    /* Decode data recieved */
    inq->deviceType = getByte(data,0);
    inq->additionalLength = getByte(data,4);
    getBytes((SANE_Byte*)(inq->vendor),data,8,8); /* Note: not 0-terminated */
    getBytes((SANE_Byte*)(inq->product),data,16,16); /* Note: not 0-terminated */
    getBytes((SANE_Byte*)(inq->productRevision),data,32,4); /* Note: not 0-terminated */
    /* 1st Vendor-specific block, 20 bytes, see pie_get_inquiry_values(), partially: */
    inq->maxResolutionX = getShort(data,36);
    inq->maxResolutionY = getShort(data,38);
    inq->maxScanWidth = getShort(data,40);
    inq->maxScanHeight = getShort(data,42);
    inq->filters = getByte(data,44);
    inq->colorDepths = getByte(data,45);
    inq->colorFormat = getByte(data,46);
    inq->imageFormat = getByte(data,48);
    inq->scanCapability = getByte(data,49);
    inq->optionalDevices = getByte(data,50);
    inq->enhancements = getByte(data,51);
    inq->gammaBits = getByte(data,52);
    inq->lastFilter = getByte(data,53);
    inq->previewScanResolution = getShort(data,54);
    /* 2nd vendor specific block (36 bytes at offset 96) */
    getBytes((SANE_Byte*)(inq->firmwareVersion),data,96,4); inq->firmwareVersion[4]=0x00;
    inq->halftones = getByte(data,100);
    inq->minumumHighlight = getByte(data,101);
    inq->maximumShadow = getByte(data,102);
    inq->calibrationEquation = getByte(data,103);
    inq->maximumExposure = getShort(data,104);
    inq->minimumExposure = getShort(data,106);
    inq->x0 = getShort(data,108);
    inq->y0 = getShort(data,110);
    inq->x1 = getShort(data,112);
    inq->y1 = getShort(data,114);
    inq->model = getShort(data,116);
    getBytes((SANE_Byte*)(inq->production),data,120,24);
    getBytes((SANE_Byte*)(inq->signature),data,144,40);
    /* remove newline in signature */
    for (k=0; k<40; k++) if (inq->signature[k]=='\n') inq->signature[k]=' ';
}

/**
 * Set scan mode parameters, such as resolution, colors to scan, color depth,
 * color format, and a couple of scan quality settings (sharpen, skip
 * calibration, fast infrared). It performs the SCSI-command MODE SELECT,
 * code 0x15.
 * 
 * @param device_number Device number
 * @param mode Mode parameters
 * @return SANE_status
 * @see Reflecta_Mode
 */
void cmdSetMode(SANE_Int device_number, struct Reflecta_Mode* mode, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    SANE_Byte command[6];
    SANE_Int size = 16;
    SANE_Byte data[size];
    
    DBG (DBG_sane_info, "cmdSetMode()\n");
    
    setCommand(command,SCSI_MODE_SELECT,size);
    
    DBG (DBG_info, "cmdSetMode() set:\n");
    DBG (DBG_info, " resolution = %d\n",mode->resolution);
    DBG (DBG_info, " passes = %02x\n",mode->passes);
    DBG (DBG_info, " depth = %02x\n",mode->colorDepth);
    DBG (DBG_info, " color format = %02x\n",mode->colorFormat);
    DBG (DBG_info, " sharpen = %d\n",mode->sharpen);
    DBG (DBG_info, " skip calibration = %d\n",mode->skipCalibration);
    DBG (DBG_info, " fast infrared = %d\n",mode->fastInfrared);
    DBG (DBG_info, " halftone pattern = %d\n",mode->halftonePattern);
    DBG (DBG_info, " line threshold = %d\n",mode->lineThreshold);

    /* Code data */
    memset(data,'\0',size);
    setByte(size-1,data,1);
    setShort(mode->resolution,data,2);
    setByte(mode->passes,data,4);
    setByte(mode->colorDepth,data,5);
    setByte(mode->colorFormat,data,6);
    setByte(mode->byteOrder,data,8);
    SANE_Byte quality = 0x00;
    if (mode->sharpen) quality |= 0x02;
    if (mode->skipCalibration) quality |= 0x08;
    if (mode->fastInfrared) quality |= 0x80;
    setByte(quality,data,9);
    setByte(mode->halftonePattern,data,12);
    setByte(mode->lineThreshold,data,13);

    commandScannerRepeat(device_number, command, data, size, status, repeat);
}

/* SCSI COPY, code 0x18 */
/**
 * Get the currently used CCD-mask, which defines which pixels have been used in
 * the scan, and which allows to relate scan data to shading data. A mask is a
 * 5340 byte array which consists only contains the values 0x00 and 0x70. A
 * value of 0x00 indicates the pixel is used, a value of 0x70 that it is not.\n
 * The number of 0x00 bytes equals the number of pixels on a line.\n
 * The mask begins with a number of 0x70 bytes equal to the scan frame x0-value
 * divided by 2.\n
 * The SCSI-command COPY (code 0x18) is used for function.
 *
 * @param device_number Device number
 * @param mask 
 * @return SANE_status
 */
void cmdGetCCDMask(SANE_Int device_number, SANE_Byte* mask, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    SANE_Byte command[6];
    SANE_Int size = 5340;
    
    DBG (DBG_sane_info, "cmdGetCCDMask()\n");
    
    setCommand(command,SCSI_COPY,size);    

    memset(mask,'\0',size);
    commandScannerRepeat(device_number, command, mask, size, status, repeat);
    
/*
    int k;
    fprintf(stderr,"CCD mask:\n");
    for (k = 0; k < size; k++) {
        fprintf(stderr,"%02x ",mask[k]);
        if ((k+1) % 50 == 0) fprintf(stderr,"\n");
    }
*/
    
}

/**
 * Get scan mode parameters, such as resolution, colors to scan, color depth,
 * color format, and a couple of scan quality settings (sharpen, skip
 * calibration, fast infrared). It performs the SCSI-command MODE SELECT,
 * code 0x1A.
 * 
 * @param device_number Device number
 * @param mode Mode parameters
 * @return SANE_status
 * @see Reflecta_Mode
 */
void cmdGetMode(SANE_Int device_number, struct Reflecta_Mode* mode, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    DBG (DBG_sane_info, "cmdGetMode(): not implemented\n");
    status->sane_status = SANE_STATUS_INVAL;
}

/**
 * Start a scan (SCSI SCAN command, code 0x1B, size byte = 0x01).\n
 * There are four phases in a scan process. During each phase a limited number of
 * commands is available. The phases are:\n
 * 1. Calibration phase: make previously collected shading correction data available\n
 * 2. Line-by-line scan & read phase\n
 * 3. Output CCD-mask phase\n
 * 4. Scan and output scan data phase\n
 * 
 * The calibration phase is skipped if Reflecta_Mode.skipCalibration is set. If
 * the scanner determines a calibration is necessary, a CHECK CONDIDITION response
 * is returned. Available command during this phase:\n
 * 1. cmdIsUnitReady()\n
 * 2. cmdGetScannedLines(): read shading correction lines\n
 * 3. cmdStopScan: abort scanning process\n
 * 4. cmdGetOptimizedSettings() : the settings are generated during the initialisation of this phase, so they are current\n
 * 5. cmdSetSettings(): settings take effect in the next scan phase\n\n
 * The line-by-line phase is only entered if Reflecta_Mode.div_10[0] bit 5 is
 * set. It is not implemented.\n\n
 * In the CCD-mask output phase the CCD-mask is read. Available command during this phase:\n
 * 1. cmdIsUnitReady()\n
 * 2. cmdGetCCDMask()\n
 * 3. cmdStopScan: abort scanning process\n\n
 * In the 'scan and output scan data' phase, the slide is scanned while data is
 * read in the mean time. Available command during this phase:\n
 * 1. cmdIsUnitReady()\n
 * 2. cmdGetScannedLines()\n
 * 2. cmdGetScanParameters()\n
 * 4. cmdStopScan: abort scanning process\n
 * 
 * @param device_number Device number
 * @return SANE_status
 */
void cmdStartScan(SANE_Int device_number, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    SANE_Byte command[6];
    
    DBG (DBG_sane_info, "cmdStartScan()\n");
    
    setCommand(command,SCSI_SCAN,1);

    commandScannerRepeat(device_number, command, NULL, 0, status, repeat);
}

/**
 * Stop a scan started with cmdStartScan(). It issues a SCSI SCAN command,
 * code 0x1B, with size byte = 0x00.
 * 
 * @param device_number Device number
 * @return SANE_status
 */
void cmdStopScan(SANE_Int device_number, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    SANE_Byte command[6];
    
    DBG (DBG_sane_info, "cmdStopScan()\n");
    
    setCommand(command,SCSI_SCAN,0);

    commandScannerRepeat(device_number, command, NULL, 0, status, repeat);
}

/**
 * Set scan head to a specific position, depending on the value for mode:\n
 * mode = 1: Returns the scan head to the resting position, after a short move
 * forward. If this command is left out between two scans, the second scan is
 * up-down-mirrored, and scanning starts where the proevious scan stopped.\n
 * mode = 2: Resets the scan head an then moves it forward depending on 'size',
 * but it is a bit unpredictable to what position. The scanner may attempt to
 * move the head past its physical end position. The mode is not implemented.\n
 * mode = 3: This command positions the scan head to the start of the slide.\n
 * mode = 4 or 5: The command forwards (4) or retreats (5) the scan head the
 * given amount of steps (in size).\n
 * The SCSI code is 0xD2, there is no related command name.
 * 
 * @param device_number Device number
 * @param mode
 * @param size
 * @return SANE_status
 */
void cmdSetScanHead(SANE_Int device_number, SANE_Int mode, SANE_Int steps, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    SANE_Byte command[6];
    SANE_Int size = 4;
    SANE_Byte data[size];
    
    DBG (DBG_sane_info, "cmdSetScanHead()\n");
    
    setCommand(command,SCSI_SET_SCAN_HEAD,size);

    /* Code data */
    memset(data,'\0',size);
    switch (mode) {
        case 1:
            data[0] = 2;
            break;
        case 2:
            DBG (DBG_sane_info, "cmdSetScanHead() mode 2 unreliable, possibly dangerous\n");
            status->sane_status = SANE_STATUS_INVAL;
            return;
        case 3:
            data[0] = 8;
            break;
        case 4:
            data[0] = 0; /* forward */
            data[2] = (steps>>8) & 0xFF;
            data[3] = steps & 0xFF;
            break;
        case 5:
            data[0] = 1; /* backward */
            data[2] = (steps>>8) & 0xFF;
            data[3] = steps & 0xFF;
            break;
    }
    
    commandScannerRepeat(device_number, command, data, size, status, repeat);
}

/**
 * Get internal scanner settings which have resulted from an auto-calibration
 * procedure. This procedure only runs when calibrating (Scan phase 1), so the
 * data returned are relatively static.\n
 * The SCSI code is 0xD7, there is no related command name.
 * 
 * @param device_number Device number
 * @param settings Settings for gain and offset for the four colors RGBI
 * @return SANE_status
 * @see Reflecta_Settings
 */
void cmdGetGainOffset(SANE_Int device_number, struct Reflecta_Settings* settings, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    SANE_Byte command[6];
    SANE_Int size = 103;
    SANE_Byte data[size];
    int k;
    SANE_Byte val[3];

    DBG (DBG_sane_info, "cmdGetOptimizedSettings()\n");
    
    setCommand(command,SCSI_READ_GAIN_OFFSET,size);    
    
    memset(data,'\0',size);
    commandScannerRepeat(device_number, command, data, size, status, repeat);
    if (status->sane_status != SANE_STATUS_GOOD) {
        return;
    }
    
    /* Decode data received */
    getShorts(settings->saturationLevel,data,54,3);
    getShorts(settings->exposureTime,data,60,3);
    getBytes(val,data,66,3);
    for (k=0; k<3; k++) settings->offset[k] = val[k];
    getBytes(val,data,72,3);
    for (k=0; k<3; k++) settings->gain[k] = val[k];
    settings->light = getByte(data,75);
    settings->exposureTime[3] = getShort(data,98);
    settings->offset[3] = getByte(data,100);
    settings->gain[3] = getByte(data,102);

    DBG (DBG_info, "cmdGetGainOffset() set:\n");
    DBG (DBG_info, " saturationlevels = %d-%d-%d\n",settings->saturationLevel[0],settings->saturationLevel[1],settings->saturationLevel[2]);
    DBG (DBG_info, " minimumExposureTime = %d\n",settings->minimumExposureTime);
    DBG (DBG_info, " ---\n");
    DBG (DBG_info, " exposure times = %d-%d-%d-%d\n",settings->exposureTime[0],settings->exposureTime[1],settings->exposureTime[2],settings->exposureTime[3]);
    DBG (DBG_info, " gain = %d-%d-%d-%d\n",settings->gain[0],settings->gain[1],settings->gain[2],settings->gain[3]);
    DBG (DBG_info, " offset = %d-%d-%d-%d\n",settings->offset[0],settings->offset[1],settings->offset[2],settings->offset[3]);
    DBG (DBG_info, " light = %02x\n",settings->light);
    DBG (DBG_info, " double times = %02x\n",settings->doubleTimes);
    DBG (DBG_info, " extra entries = %02x\n",settings->extraEntries);

}


/**
 * Set internal scanner settings such as gain and offset.\n
 * There are two effective moments for this command:\n
 * 1. For a scan without calibration phase: before the cmdStartScan() command;
 * 2. For a sccan with calibration phase: before (or during) reading the shading reference data.
 * The SCSI code is 0xDC, there is no related command name.
 * 
 * @param device_number Device number
 * @param settings Settings for gain and offset for the four colors RGBI
 * @return SANE_status
 * @see Reflecta_Settings
 */
void cmdSetGainOffset(SANE_Int device_number, struct Reflecta_Settings* settings, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    SANE_Byte command[6];
    SANE_Int size = 23;
    SANE_Byte data[size];
    int k;
    SANE_Byte val[3];
    
    DBG (DBG_sane_info, "cmdSetGainOffset()\n");
    
    setCommand(command,SCSI_WRITE_GAIN_OFFSET,size);

    DBG (DBG_info, "cmdSetGainOffset() set:\n");
    DBG (DBG_info, " exposure times = %d-%d-%d-%d\n",settings->exposureTime[0],settings->exposureTime[1],settings->exposureTime[2],settings->exposureTime[3]);
    DBG (DBG_info, " gain = %d-%d-%d-%d\n",settings->gain[0],settings->gain[1],settings->gain[2],settings->gain[3]);
    DBG (DBG_info, " offset = %d-%d-%d-%d\n",settings->offset[0],settings->offset[1],settings->offset[2],settings->offset[3]);
    DBG (DBG_info, " light = %02x\n",settings->light);
    DBG (DBG_info, " double times = %02x\n",settings->doubleTimes);
    DBG (DBG_info, " extra entries = %02x\n",settings->extraEntries);
   
    /* Code data */
    memset(data,'\0',size);
    setShorts(settings->exposureTime,data,0,3);
    for (k=0; k<3; k++) val[k] = settings->offset[k];
    setBytes(val,data,6,3);
    for (k=0; k<3; k++) val[k] = settings->gain[k];
    setBytes(val,data,12,3);
    setByte(settings->light,data,15);
    setByte(settings->extraEntries,data,16);
    setByte(settings->doubleTimes,data,17);
    setShort(settings->exposureTime[3],data,18);
    setByte(settings->offset[3],data,20);
    setByte(settings->gain[3],data,22);
    
    commandScannerRepeat(device_number, command, data, size, status, repeat);
}

/**
 * Get scanner state information: button pushed,
 * warming up, scanning.
 * The SCSI code is 0xDD, there is no related command name.
 * 
 * @param device_number Device number
 * @param state State information
 * @return SANE_status
 */
void cmdGetState(SANE_Int device_number, struct Reflecta_Scanner_State* state, struct Reflecta_Command_Status *status, SANE_Int repeat)
{
    SANE_Byte command[6];
    SANE_Byte data[11];
    SANE_Int size = 11;
    
    /* Execute READ STATUS command */
    DBG (DBG_sane_info, "cmdGetState()\n");
    
    setCommand(command,SCSI_READ_STATE,size);
    
    memset(data,'\0',size);
    commandScannerRepeat(device_number, command, data, size, status, repeat);
    if (status->sane_status != SANE_STATUS_GOOD) {
        return;
    }

    /* Decode data recieved */
    state->buttonPushed = getByte(data,0);
    state->warmingUp = getByte(data,5);
    state->scanning = getByte(data,6);
}

/**
 * Prepare 6-byte command array with command code and size value
 * 
 * @param command
 * @param code
 * @param size
 */
void setCommand(SANE_Byte* command, SANE_Byte code, SANE_Word size)
{
    memset(command,'\0',6);
    command[0] = code;
    command[3] = (size>>8) & 0xFF;
    command[4] = size & 0xFF;
}
